package sample.resource;

import java.net.URI;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import javax.enterprise.context.RequestScoped;
import javax.inject.Inject;
import javax.ws.rs.Consumes;
import javax.ws.rs.ForbiddenException;
import javax.ws.rs.FormParam;
import javax.ws.rs.GET;
import javax.ws.rs.NotFoundException;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriBuilder;
import javax.ws.rs.core.UriInfo;

import sample.entity.Account;
import sample.entity.Role;
import sample.entity.Shinsei;
import sample.entity.Status;
import sample.security.Protected;
import sample.security.User;
import sample.service.ShinseiService;
import sample.ui.Jsp;

/**
 * 申請に関するHTTPリクエストのエンドポイントとなるリソースクラスです。
 *
 */
@Protected
@RequestScoped
@Path("shinsei")
public class ShinseiResource {

    @Inject
    private ShinseiService service;
    @Inject
    private User user;

    /**
     * すべての申請データを表示します。
     * 
     * @return JSPへフォワードするためのオブジェクト
     */
    @GET
    @Produces(MediaType.TEXT_HTML)
    public Jsp findAll() {
        List<Shinsei> entities = service.findAll();
        ListModel model = new ListModel(entities);
        return new Jsp("/WEB-INF/views/shinseiList.jsp", model);
    }

    /**
     * 申請データの新規登録画面を開きます。
     * 
     * @return JSPへフォワードするためのオブジェクト
     */
    @Path("new")
    @GET
    @Produces(MediaType.TEXT_HTML)
    public Jsp openBlank() {

        //新規登録できるのは申請者のみ
        getUserAsApplicant();

        return new Jsp("/WEB-INF/views/newShinsei.jsp");
    }

    /**
     * 申請データを作成します。
     * 作成された申請データのステータスは「未申請」です。
     * 
     * @param description 概要
     * @param detail 詳細
     * @param uriInfo UriInfoインスタンス
     * @return 作成された申請データ
     */
    @POST
    @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
    public Response create(@FormParam("description") String description,
            @FormParam("detail") String detail,
            @Context UriInfo uriInfo) {

        //申請できるのは申請者のみ
        getUserAsApplicant();

        Shinsei shinsei = service.create(description, detail);

        //リダイレクトURIを組み立てる
        UriBuilder uriBuilder = uriInfo.getRequestUriBuilder();
        URI location = uriBuilder.path("{id}").build(shinsei.getId());

        return Response.created(location).build();
    }

    /**
     * リクエストパスからIDを取り出して、サブリソースを返します。
     * 
     * @param id ID
     * @return サブリソース
     */
    @Path("{id}")
    public SubApi get(@PathParam("id") Long id) {
        return new SubApi(id);
    }

    /**
     * 申請のサブリソースです。
     *
     */
    @Protected
    public class SubApi {

        /**
         * ID
         */
        private final Long id;

        /**
         * IDを設定したインスタンスを構築します。
         * 
         * @param id ID
         */
        public SubApi(Long id) {
            this.id = id;
        }

        /**
         * 指定されたIDの申請データを表示します。
         * 
         * @return JSPへフォワードするためのオブジェクト
         */
        @GET
        @Produces(MediaType.TEXT_HTML)
        public Jsp findById() {
            Shinsei entity = service.findById(id)
                    .orElseThrow(NotFoundException::new);
            SingleModel model = new SingleModel(entity);
            return new Jsp("/WEB-INF/views/shinsei.jsp", model);
        }

        /**
         * 指定されたIDの申請データを申請します。
         * 
         * @return 申請された申請データ
         */
        @Path("apply")
        @POST
        @Produces(MediaType.APPLICATION_JSON)
        public Shinsei apply() {
            Account applicant = getUserAsApplicant();
            return service.apply(id, applicant)
                    .orElseThrow(NotFoundException::new);
        }

        /**
         * 指定された申請データを承認します。
         * 
         * @param reason 承認理由
         * @return 承認された申請データ
         */
        @Path("approve")
        @POST
        @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
        @Produces(MediaType.APPLICATION_JSON)
        public Shinsei approve(@FormParam("reason") String reason) {
            Account approver = getUserAsApprover();
            return service.approve(id, approver, reason)
                    .orElseThrow(NotFoundException::new);
        }

        /**
         * 指定された申請データを棄却します。
         * 
         * @param reason 棄却理由
         * @return 棄却された申請データ
         */
        @Path("reject")
        @POST
        @Consumes(MediaType.APPLICATION_FORM_URLENCODED)
        @Produces(MediaType.APPLICATION_JSON)
        public Shinsei reject(@FormParam("reason") String reason) {
            Account approver = getUserAsApprover();
            return service.reject(id, approver, reason)
                    .orElseThrow(NotFoundException::new);
        }
    }

    /**
     * 申請一覧画面に渡す値をまとめたクラスです。
     *
     */
    public static class ListModel {

        /**
         * 申請データのリスト
         */
        private final List<Shinsei> list;

        /**
         * 申請データのリストを設定したインスタンスを構築します。
         * 
         * @param list 申請データのリスト
         */
        public ListModel(List<Shinsei> list) {
            this.list = Collections.unmodifiableList(list);
        }

        /**
         * 申請データのリストを返します。
         * 
         * @return 申請データのリスト
         */
        public List<Shinsei> getList() {
            return list;
        }

        /**
         * 指定されたアカウントが申請データを作成可能か否かを返します。
         * 
         * @param account 対象アカウント
         * @return 申請データを作成可能ならtrue
         */
        public boolean isEnabledToCreateBy(Account account) {
            return account.getRole() == Role.APPLICANT;
        }
    }

    /**
     * 申請詳細画面に渡す値をまとめたクラスです。
     *
     */
    public static class SingleModel {

        /**
         * 申請データ
         */
        private final Shinsei entity;

        /**
         * 申請データを設定したインスタンスを構築します。
         * 
         * @param entity 申請データ
         */
        public SingleModel(Shinsei entity) {
            this.entity = Objects.requireNonNull(entity);
        }

        /**
         * 申請データを返します。
         * 
         * @return 申請データ
         */
        public Shinsei getEntity() {
            return entity;
        }

        /**
         * 申請済みか否かを返します。
         * 
         * @return 申請済みならtrue
         */
        public boolean isApplied() {
            return entity.getStatus() != Status.YET;
        }

        /**
         * 承認済みか否かを返します。
         * 
         * @return 承認済みならtrue
         */
        public boolean isApproved() {
            return entity.getStatus() == Status.APPROVED;
        }

        /**
         * 棄却済みか否かを返します。
         * 
         * @return 棄却済みならtrue
         */
        public boolean isRejected() {
            return entity.getStatus() == Status.REJECTED;
        }

        /**
         * 指定されたアカウントが申請可能か否かを返します。
         * 
         * @param account 対象アカウント
         * @return 申請可能ならtrue
         */
        public boolean isEnabledToApplyBy(Account account) {
            return entity.getStatus() == Status.YET && account.getRole() == Role.APPLICANT;
        }

        /**
         * 指定されたアカウントが承認(棄却)可能か否かを返します。
         * 
         * @param account 対象アカウント
         * @return 承認(棄却)可能ならtrue
         */
        public boolean isEnabledToApproveBy(Account account) {
            return entity.getStatus() == Status.APPLYING && account.getRole() == Role.APPROVER;
        }
    }

    private Account getUserAsApplicant() {
        return user.getAccount()
                .filter(a -> a.getRole() == Role.APPLICANT)
                .orElseThrow(ForbiddenException::new);
    }

    private Account getUserAsApprover() {
        return user.getAccount()
                .filter(a -> a.getRole() == Role.APPROVER)
                .orElseThrow(ForbiddenException::new);
    }
}
