package sample.service;

import java.time.LocalDateTime;
import java.util.List;
import java.util.Optional;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.transaction.Transactional;

import sample.entity.Account;
import sample.entity.Shinsei;
import sample.entity.Status;
import sample.exception.IllegalStatusException;

/**
 * 申請に関するビジネスロジックをまとめたクラスです。
 *
 */
@ApplicationScoped
//このクラスのメソッドをトランザクション境界にすることを表すアノテーションです。
@Transactional
public class ShinseiService {

    //本誌では @PersistenceContext でインジェクションしていましたが、
    //EntityManagerもCDI管理ビーンにして @Inject でインジェクション
    //する方が柔軟に扱えます。
    //EntityManagerをCDI管理ビーンにする方法は JpaProvider クラスを
    //確認してください。
    @Inject
    private EntityManager em;

    /**
     * すべての申請データを抽出します。
     * 
     * @return 申請データのリスト
     */
    public List<Shinsei> findAll() {
        String query = "SELECT a FROM Shinsei a";
        List<Shinsei> entities = em.createQuery(query, Shinsei.class).getResultList();
        return entities;
    }

    /**
     * 指定されたIDで申請データを検索します。
     * 
     * @param id ID
     * @return 申請データ
     */
    public Optional<Shinsei> findById(Long id) {
        Shinsei entity = em.find(Shinsei.class, id);
        return Optional.ofNullable(entity);
    }

    /**
     * 新しく申請データを作成します。
     * 
     * @param description 概要
     * @param detail 詳細
     * @return 作成された申請データ
     */
    public Shinsei create(String description, String detail) {
        Shinsei entity = new Shinsei();
        entity.setDescription(description);
        entity.setDetail(detail);
        entity.setStatus(Status.YET);
        em.persist(entity);
        return entity;
    }

    /**
     * 指定されたIDに該当する申請データのステータスを「申請中」に更新します。
     * 
     * @param id ID
     * @param applicant 申請者
     * @return 更新された申請データ
     */
    public Optional<Shinsei> apply(Long id, Account applicant) {
        Optional<Shinsei> opt = findById(id);
        opt.ifPresent(entity -> {
            //未申請のものが対象
            if (entity.getStatus() != Status.YET) {
                throw new IllegalStatusException();
            }
            entity.setApplicant(applicant);
            entity.setAppliedAt(LocalDateTime.now());
            entity.setStatus(Status.APPLYING);
        });
        return opt;
    }

    /**
     * 指定されたIDに該当する申請データのステータスを「承認」に更新します。
     * 
     * @param id ID
     * @param approver 承認者
     * @param reason 承認理由
     * @return 更新された申請データ
     */
    public Optional<Shinsei> approve(Long id, Account approver, String reason) {
        Optional<Shinsei> opt = findById(id);
        opt.ifPresent(entity -> {
            //申請中のものが対象
            if (entity.getStatus() != Status.APPLYING) {
                throw new IllegalStatusException();
            }
            entity.setApprover(approver);
            entity.setApprovedAt(LocalDateTime.now());
            entity.setReason(reason);
            entity.setStatus(Status.APPROVED);
        });
        return opt;
    }

    /**
     * 指定されたIDに該当する申請データのステータスを「棄却」に更新します。
     * 
     * @param id ID
     * @param approver 棄却者
     * @param reason 棄却理由
     * @return 更新された申請データ
     */
    public Optional<Shinsei> reject(Long id, Account approver, String reason) {
        Optional<Shinsei> opt = findById(id);
        opt.ifPresent(entity -> {
            //申請中のものが対象
            if (entity.getStatus() != Status.APPLYING) {
                throw new IllegalStatusException();
            }
            entity.setApprover(approver);
            entity.setApprovedAt(LocalDateTime.now());
            entity.setReason(reason);
            entity.setStatus(Status.REJECTED);
        });
        return opt;
    }
}
