package sample.ui;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.nio.charset.StandardCharsets;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;
import javax.ws.rs.InternalServerErrorException;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.ext.MessageBodyWriter;
import javax.ws.rs.ext.Provider;

import sample.security.User;

/**
 * JAX-RSとJSPを連携させるためのクラスです。
 *
 */
//JAX-RSにこのクラスを認識させるためのアノテーションです。
@Provider
@ApplicationScoped
public class JspProvider implements MessageBodyWriter<Jsp> {

    //JAX-RSにインジェクションさせるためのアノテーションです
    @Context
    private HttpServletRequest request;
    @Context
    private HttpServletResponse response;

    @Inject
    private User user;

    @Override
    public boolean isWriteable(Class<?> type, Type genericType, Annotation[] annotations,
            MediaType mediaType) {
        return true;
    }

    @Override
    public long getSize(Jsp t, Class<?> type, Type genericType, Annotation[] annotations,
            MediaType mediaType) {
        return -1;
    }

    @Override
    public void writeTo(Jsp t, Class<?> type, Type genericType, Annotation[] annotations,
            MediaType mediaType, MultivaluedMap<String, Object> httpHeaders,
            OutputStream entityStream) throws IOException, WebApplicationException {
        request.setAttribute("model", t.getModel());
        user.getAccount().ifPresent(x -> request.setAttribute("user", x));
        RequestDispatcher dispatcher = request.getRequestDispatcher(t.getPath());
        try (PrintWriter writer = new NeverClosedPrintWriter(entityStream)) {
            HttpServletResponse wrappedResponse = new WriterBoundHttpServletResponse(response,
                    writer);
            dispatcher.forward(request, wrappedResponse);

        } catch (ServletException e) {
            throw new InternalServerErrorException(e);
        }
    }

    /**
     * OutputStreamをcloseしないためのラッパークラスです。
     *
     */
    private static class NeverClosedPrintWriter extends PrintWriter {

        public NeverClosedPrintWriter(OutputStream out) {
            super(new OutputStreamWriter(out, StandardCharsets.UTF_8));
        }

        @Override
        public void close() {
            flush();
        }
    }

    /**
     * getWriterメソッドで指定されたPrintWriterを返すためのラッパークラスです。
     *
     */
    private static class WriterBoundHttpServletResponse extends HttpServletResponseWrapper {

        private final PrintWriter writer;

        public WriterBoundHttpServletResponse(HttpServletResponse response, PrintWriter writer) {
            super(response);
            this.writer = writer;

        }

        @Override
        public PrintWriter getWriter() throws IOException {
            return writer;
        }
    }
}
